"""Core features include an init function for skeleton"""

import logging
import re
from types import ModuleType
from typing import Any

import click
from click_help_colors.utils import _colorize
from dotmap import DotMap  # type: ignore

from click_skeleton.advanced_group import AdvancedGroup
from click_skeleton.decorators import add_options
from click_skeleton.defaults import DEFAULT_CONTEXT_SETTINGS
from click_skeleton.version import version_cmd

logger = logging.getLogger(__name__)


def version_option(
    version: None | str = None,
    prog_name: None | str = None,
    message: str = "%(prog)s, version %(version)s",
    message_color: None | str = None,
    prog_name_color: None | str = None,
    version_color: None | str = None,
    **kwargs: Any,
) -> Any:
    """Re-implement version handling with --version and -V shortcut"""
    msg_parts = []
    if prog_name is None:
        prog_name = ""
    if version is None:
        version = ""
    for placeholder in re.split(r"(%\(version\)s|%\(prog\)s)", message):
        if placeholder == "%(prog)s":
            msg_parts.append(_colorize(prog_name, prog_name_color or message_color))
        elif placeholder == "%(version)s":
            msg_parts.append(_colorize(version, version_color or message_color))
        else:
            msg_parts.append(_colorize(placeholder, message_color))
    message = "".join(msg_parts)

    return click.version_option(
        version, "--version", "-V", prog_name=prog_name, message=message, **kwargs
    )


def skeleton(
    name: str,
    version: str,
    auto_envvar_prefix: None | str = None,
    cls: Any = None,
    commands: None | dict[str, Any] = None,
    groups_package: None | ModuleType = None,
    **kwargs: Any,
) -> Any:
    """Generates an skeleton group with version options included"""
    auto_envvar_prefix = (
        auto_envvar_prefix if auto_envvar_prefix is not None else name.upper()
    )
    if cls is None:
        cls = AdvancedGroup

    sensible_context_settings = DEFAULT_CONTEXT_SETTINGS
    sensible_context_settings["auto_envvar_prefix"] = auto_envvar_prefix

    obj = DotMap()
    obj.prog_name = name
    obj.version = version
    obj.context_settings = sensible_context_settings

    context_settings = sensible_context_settings
    context_settings["auto_envvar_prefix"] = auto_envvar_prefix
    context_settings["obj"] = obj

    commands = commands if commands is not None else {}
    commands["version"] = version_cmd
    if groups_package is not None and cls is AdvancedGroup:
        kwargs["groups_package"] = groups_package

    return add_options(
        version_option(
            version=version,
            prog_name=name,
            version_color="green",
            prog_name_color="yellow",
        ),
        click.group(
            name=name,
            context_settings=context_settings,
            commands=commands,
            cls=cls,
            **kwargs,
        ),
    )



"""Generates a readme composed of all commands and subcommands help strings"""

from textwrap import indent
from typing import Any, Dict

import click

from . import AdvancedGroup


def write_header(text: str, char: str) -> None:
    """Write line followed by a newline containing as many chars"""
    click.echo("\n" + text)
    click.echo(char * len(text))


def write_codeblock(code: str, output: str) -> None:
    """Write a console codeblock, either in rst or markdown"""
    if output == "rst":
        indented_code = indent(code, "  ")
        click.echo(".. code-block::\n")
        click.echo(indented_code)
    elif output == "markdown":
        click.echo("```")
        click.echo(code)
        click.echo("```")


def readme(
    main_cli: click.core.Group,
    prog_name: str,
    context_settings: Dict[str, Any],
    output: str = "rst",
) -> Any:
    """Recursively output a beautiful readme on stdout, supports 3 levels of subcommands"""
    if output == "rst":
        write_header("Commands", "-")
    elif output == "markdown":
        click.echo("# Commands\n")

    base_ctx = click.Context(main_cli, info_name=prog_name, **context_settings)
    with base_ctx.scope():
        cli_help = main_cli.get_help(base_ctx)
        write_codeblock(cli_help, output)

        for command_name, command in sorted(main_cli.commands.items()):
            if command.hidden:
                continue

            command_header = f"{prog_name} {command_name}"
            if output == "rst":
                write_header(command_header, "*")
            elif output == "markdown":
                click.echo(f"\n## {command_header}\n")

            command_ctx = click.Context(
                command, info_name=command_name, parent=base_ctx
            )
            with command_ctx.scope():
                command_help = command.get_help(command_ctx)
                write_codeblock(command_help, output)

            if not isinstance(command, AdvancedGroup):
                continue

            for subcommand_name, subcommand in sorted(command.commands.items()):
                if subcommand_name == "help" or subcommand.hidden:
                    continue

                subcommand_header = f"{prog_name} {command_name} {subcommand_name}"
                if output == "rst":
                    write_header(f"{subcommand_header}", "*")
                elif output == "markdown":
                    click.echo(f"\n### {subcommand_header}\n")

                subcommand_ctx = click.Context(
                    subcommand, info_name=subcommand_name, parent=command_ctx
                )
                with subcommand_ctx.scope():
                    subcommand_help = subcommand.get_help(subcommand_ctx)
                    write_codeblock(subcommand_help, output)



"""Helpers tests"""

import pytest
from beartype import beartype

from click_skeleton.helpers import raise_limits, random_password, str2bool


@beartype
def test_str2bool() -> None:
    """Test that we are parsing bool string correctly"""
    assert str2bool("Y")
    assert not str2bool("N")
    with pytest.raises(Exception):
        str2bool("whatever")


@beartype
def test_password() -> None:
    """Two random password has very high probability to be different"""
    assert random_password() != random_password()


@beartype
def test_raise_limits() -> None:
    """Important helper as we may need a lot of file handles, does not work on Mac OS X"""
    assert raise_limits() is True



"""A subgroup example"""

import click

from click_skeleton import AdvancedGroup
from click_skeleton.helpers import split_arguments


@click.group(
    short_help="A sub group",
    cls=AdvancedGroup,
    aliases=["subgroup-alias"],
    invoke_without_command=True,
)
def cli() -> None:
    """I am a subgroup!"""


@cli.command(short_help="A sub command")
@click.option(
    "--myoptions",
    help="A splitted option",
    multiple=True,
    callback=split_arguments,
)
def subcommand(myoptions: list[str]) -> None:
    """I am a subcommand!"""
    print(f"hello from subcommand! {myoptions}")



"""Advanced group is a colored, what-did-you-mean with aliases commands group"""

import importlib
import logging
import pkgutil
import sys
from types import ModuleType
from typing import Any

import click
from click_aliases import ClickAliasedGroup
from click_didyoumean import DYMGroup  # type: ignore
from click_help_colors import HelpColorsGroup

from click_skeleton.exceptions import AlreadyRegistered

logger = logging.getLogger(__name__)


class AdvancedGroup(ClickAliasedGroup, DYMGroup, HelpColorsGroup):  # type: ignore[misc]
    """Special click group with default plugins enabled :
    - did-you-mean
    - click aliases for commands
    - colored help message
    - auto help command
    """

    def __init__(
        self,
        *args: Any,
        aliases: None | list[str] = None,
        groups_package: None | ModuleType = None,
        **kwargs: Any,
    ):
        kwargs["help_headers_color"] = "yellow"
        kwargs["help_options_color"] = "green"
        super().__init__(*args, **kwargs)
        self.aliases = aliases if aliases is not None else []

        @click.command("help")
        @click.pass_context
        def _help(ctx: None | click.Context) -> None:
            """Print help"""
            # we accept many commands, but only show help of the first one
            if not ctx:
                raise RuntimeError("no click context available")
            if not ctx.parent:
                raise RuntimeError("no click context parent available")
            click.echo(ctx.parent.get_help(), color=ctx.color)
            ctx.exit()

        self.add_command(_help, "help")
        if groups_package is not None:
            self.add_groups_from_package(groups_package)

    def add_group(self, cmd: "AdvancedGroup", name: str) -> None:
        """Registers another :class:`Group` with this group.  If the name
        is not provided, the name of the group is used.
        """
        self.add_command(cmd, name)

        if not cmd.aliases:
            return

        if name in self._commands:
            raise AlreadyRegistered(f"{name} group is already registered")

        self._commands[name] = cmd.aliases
        for alias in cmd.aliases:
            if alias in self._aliases:
                raise AlreadyRegistered(
                    f"Alias {alias} is already used by {self._aliases[alias]}"
                )
            self._aliases[alias] = name

    def add_groups_from_package(self, package: ModuleType) -> None:
        """loads commands dynamically, not supported by pyinstaller"""
        package_name = package.__name__
        package = sys.modules[package_name]
        modules_by_name = {
            name: importlib.import_module(package_name + "." + name)
            for loader, name, is_pkg in pkgutil.walk_packages(package.__path__)
        }
        for module_name, module in modules_by_name.items():
            try:
                cli = getattr(module, "cli")
                if cli.name != "cli":
                    self.add_group(cli, cli.name)
                else:
                    self.add_group(cli, module_name)
            except AttributeError:
                click.secho(
                    f"""Command module {module_name} does not contain a 'cli' definition""",
                    fg="red",
                    err=True,
                )



"""ExpandedPath is a improved click.Path with HOME expanded and check for write access"""

import logging
import os

import click

logger = logging.getLogger(__name__)


class ExpandedPath(click.ParamType):
    """Re-implement click Path to auto-expand "~" on UNIX"""

    envvar_list_splitter = os.path.pathsep

    def __init__(
        self,
        exists: bool = False,
        file_okay: bool = True,
        dir_okay: bool = True,
        writable: bool = False,
        readable: bool = True,
        resolve_path: bool = False,
        allow_dash: bool = False,
        path_type: str | None = None,
    ):
        self.exists = exists
        self.file_okay = file_okay
        self.dir_okay = dir_okay
        self.writable = writable
        self.readable = readable
        self.resolve_path = resolve_path
        self.allow_dash = allow_dash
        self.type = path_type

        if self.file_okay and not self.dir_okay:
            self.name = "file"
            self.path_type = "File"
        elif self.dir_okay and not self.file_okay:
            self.name = "directory"
            self.path_type = "Directory"
        else:
            self.name = "path"
            self.path_type = "Path"

    def convert(
        self, value: str, param: click.Parameter | None, ctx: click.Context | None
    ) -> str:
        """Expands user HOME and permits to check for write access without file needs to exist"""
        value = os.path.expanduser(value)
        is_dash = self.file_okay and self.allow_dash and value in (b"-", "-")
        if is_dash:
            return value

        if self.resolve_path:
            value = os.path.realpath(value)

        real_exists = os.path.exists(value)
        if self.exists and not real_exists:
            self.fail(f"{self.path_type} {value} does not exist.", param, ctx)

        if not self.file_okay and os.path.isfile(value):
            self.fail(f"{self.path_type} {value} is a file.", param, ctx)

        if not self.dir_okay and os.path.isdir(value):
            self.fail(f"{self.path_type} {value} is a directory.", param, ctx)

        if self.writable and not os.access(value, os.W_OK):
            pdir = os.path.dirname(value)
            if not pdir:
                pdir = "."
            if not os.access(pdir, os.W_OK):
                self.fail(f"{self.path_type} {value} is not writable.", param, ctx)

        if self.readable and not os.access(value, os.R_OK):
            pdir = os.path.dirname(value)
            if not pdir:
                pdir = "."
            if not os.access(pdir, os.R_OK):
                self.fail(f"{self.path_type} {value} is not readable.", param, ctx)

        return value



"""Decorators and helpers to add options to groups and commands, and compose decorators"""

from typing import Any, Iterable

import click


def flatten(iterables: Any | Iterable[Any]) -> Iterable[Any]:
    """Recursively flatten argument"""
    for element in iterables:
        if isinstance(element, Iterable) and not isinstance(element, (str, bytes)):
            yield from flatten(element)
        else:
            yield element


def add_options(*options: Any) -> Any:
    """Helps to add options to command in the form of list"""

    def _add_options(func: Any) -> Any:
        for option in reversed(list(flatten(options))):
            func = option(func)
        return func

    return _add_options


def group(name: str | None = None, **attrs: Any) -> Any:
    """Creates a new :class:`Group` with a function as callback.  This
    works otherwise the same as :func:`command` just that the `cls`
    parameter is set to :class:`Group`.
    """
    from click_skeleton.advanced_group import AdvancedGroup

    _ = attrs.setdefault("cls", AdvancedGroup)
    return click.command(name, **attrs)



"""Click skeleton helpers to ease CLI definitions"""

import logging

import click

from click_skeleton.advanced_group import AdvancedGroup
from click_skeleton.core import skeleton
from click_skeleton.decorators import add_options
from click_skeleton.expanded_path import ExpandedPath

logger = logging.getLogger(__name__)

# little hacky but prevent click from rewraping
click.formatting.HelpFormatter.write_dl.__defaults__ = (50, 2)

__all__ = [
    "AdvancedGroup",
    "ExpandedPath",
    "skeleton",
    "add_options",
]



#!/usr/bin/env python3
import click
from click_skeleton import backtrace

backtrace.hook(strip_path=False, enable_on_envvar_only=False, on_tty=False)

CONTEXT_SETTINGS = {
    'max_content_width': 140,
    'terminal_width': 140,
    'help_option_names': ['-h', '--help'],
}


@click.group("example", context_settings=CONTEXT_SETTINGS)
def cli():
    """Simple CLI example"""
    click.echo("I'm a CLI")


@cli.command()
def command():
    click.echo("I'm a command")


@cli.command()
def abort():
    raise Exception("I'm an exception!")


@cli.group()
def subgroup():
    click.echo("I'm a subgroup")


@subgroup.command()
def subcommand():
    click.echo("I'm a subcommand")


if __name__ == '__main__':
    cli()



"""Basic helpers used in many CLI"""

import datetime
import logging
import platform
import random
import re
import string
from collections import defaultdict
from typing import Any

import click

logger = logging.getLogger(__name__)
true_values = ("enabled", "y", "yes", "t", "true", "True", "on", "1")
false_values = ("", "none", "disabled", "n", "no", "f", "false", "False", "off", "0")


def split_arguments(
    ctx: click.Context,
    param: click.Parameter,
    value: Any,
    clean: bool = True,
) -> list[str]:
    """Arguments can be comma separated"""
    if not param.name:
        logger.error("no option name")
        raise click.Abort()

    ctx.params[param.name] = []
    for maybe_argument_list in value:
        for argument in mysplit(maybe_argument_list, ","):
            if argument in ctx.params[param.name]:
                click.secho(
                    f"{argument} was provided too many times", fg="yellow", err=True
                )
                continue
            if clean:
                ctx.params[param.name].append(clean_quotes(argument))
            else:
                ctx.params[param.name].append(argument)
    return ctx.params[param.name]


def clean_quotes(text: str) -> str:
    """Clean argument"""
    return text.replace('"', "").replace("'", "")


def recursive_str(data: Any) -> Any:
    """Recursively convert multiple data type to string"""
    if isinstance(data, list):
        return [recursive_str(x) for x in data]
    if isinstance(data, str):
        return data
    if isinstance(data, dict):
        return {recursive_str(key): recursive_str(val) for key, val in data.items()}
    return str(data)


def str2bool(val: Any) -> bool:
    """Converts any value to string and detects if it looks like a known bool value"""
    val = str(val).lower()
    if val in true_values:
        return True
    if val in false_values:
        return False
    raise ValueError(f"invalid truth value {val}")


def str_is_true(text: str) -> bool:
    """Detects a known truthful string"""
    return str(text).lower() in true_values


def str_is_false(text: str) -> bool:
    """Detects a known falsy string"""
    return str(text).lower() in false_values


def seconds_to_human(seconds: int) -> str:
    """Human readable duration from seconds"""
    return str(datetime.timedelta(seconds=seconds))


def strip_colors(text: str) -> str:
    """Remove all ANSI control characters"""
    ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    return ansi_escape.sub("", text)


def mysplit(text: str, delimiter: str = ",", **kwargs: Any) -> list[str]:
    """Split a string without returning empty list if string is empty"""
    return [x for x in text.split(sep=delimiter, **kwargs) if x]


def raise_limits() -> bool:
    """Permits to open a lot of system handles, supports Windows"""
    if "Windows" in platform.system():
        logger.warning("Cannot raise system limits on Windows")
        return True

    if "Darwin" in platform.system():
        logger.warning("Cannot raise system limits on Darwin")
        return True

    if "macOS" in platform.system():
        logger.warning("Cannot raise system limits on macOS")
        return True

    # resource module is only available on UNIX systems
    import resource  # pylint: disable=import-error

    try:
        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        logger.info(f"Current limits, soft and hard : {soft} {hard}")
        resource.setrlimit(resource.RLIMIT_NOFILE, (hard, hard))
        return True
    except (ValueError, OSError) as error:
        if "macOS" in platform.platform():
            logger.info(
                f"Cannot raise system limits for current user on : {platform.platform()}"
            )
        else:
            logger.critical(f"You may need to check ulimit parameter: {error}")
    return False


def random_password(size: int = 8) -> str:
    """Generates a random string from size, composed of ascii letters and digits"""
    alphabet = string.ascii_letters + string.digits
    return "".join(random.choice(alphabet) for i in range(size))


class PrettyDefaultDict(defaultdict):
    """Beautiful (correct printing) dict using collections.defaultdict"""

    __repr__ = dict.__repr__



# pylint: disable=missing-module-docstring,missing-function-docstring
import logging
from typing import Any

import pytest
from beartype import beartype

from click_skeleton.helpers import strip_colors
from click_skeleton.testing import run_cli
from examples.cli import __version__, main_cli

logger = logging.getLogger(__name__)


@beartype
def test_cli(cli_runner: Any) -> None:
    run_cli(cli_runner, main_cli)


@beartype
def test_version(cli_runner: Any) -> None:
    output1 = strip_colors(run_cli(cli_runner, main_cli, ["-V"]))
    output2 = strip_colors(run_cli(cli_runner, main_cli, ["--version"]))
    output3 = strip_colors(run_cli(cli_runner, main_cli, ["version"]))
    assert output1 == output2 == output3
    assert __version__ in output1


@beartype
def test_help(cli_runner: Any) -> None:
    output1 = strip_colors(run_cli(cli_runner, main_cli, ["-h"]))
    output2 = strip_colors(run_cli(cli_runner, main_cli, ["--help"]))
    output3 = strip_colors(run_cli(cli_runner, main_cli, ["help"]))
    assert output1 == output2 == output3


@beartype
def test_subcommand_help(cli_runner: Any) -> None:
    output1 = strip_colors(run_cli(cli_runner, main_cli, ["subgroup", "-h"]))
    output2 = strip_colors(run_cli(cli_runner, main_cli, ["subgroup", "--help"]))
    output3 = strip_colors(run_cli(cli_runner, main_cli, ["subgroup", "help"]))
    assert output1 == output2 == output3


@beartype
def test_subcommand_raise(cli_runner: Any) -> None:
    with pytest.raises(Exception):
        run_cli(cli_runner, main_cli, ["abort"])


@beartype
def test_subcommand_expanded_path(cli_runner: Any) -> None:
    run_cli(cli_runner, main_cli, ["expanded-path"])


@beartype
def test_readme_rst(cli_runner: Any) -> None:
    run_cli(cli_runner, main_cli, ["readme", "--output", "rst"])


@beartype
def test_readme_markdown(cli_runner: Any) -> None:
    run_cli(cli_runner, main_cli, ["readme", "--output", "markdown"])


@beartype
def test_subgroup(cli_runner: Any) -> None:
    run_cli(cli_runner, main_cli, ["subgroup"])


@beartype
def test_subcommand(cli_runner: Any) -> None:
    run_cli(
        cli_runner,
        main_cli,
        [
            "subgroup",
            "subcommand",
            "--myoptions",
            "plop1,plop2",
            "--myoptions",
            "plop3",
        ],
    )



"""Preconfigured version command for root AdvancedGroup (skeleton)"""

import click
from click_help_colors import HelpColorsCommand


@click.command(
    short_help="Print version",
    cls=HelpColorsCommand,
    help_headers_color="yellow",
    help_options_color="green",
)
@click.pass_context
def version_cmd(ctx: click.Context) -> None:
    """Print version, equivalent to -V and --version"""
    click.echo(
        f"{click.style(ctx.obj.prog_name, fg='yellow')}, version {click.style(ctx.obj.version, fg='green')}",
        color=ctx.color,
    )



"""Improved click.testing test runner"""

import logging
import traceback
from typing import Any

import click
from click import testing

from click_skeleton import helpers

logger = logging.getLogger(__name__)


def run_cli(
    cli_runner: testing.CliRunner,
    called_cli: click.core.Group,
    *args: Any,
    **kwargs: Any,
) -> str:
    """Helper to run the CLI for pytest-click unit tests"""
    result: testing.Result = cli_runner.invoke(
        called_cli,
        *args,
        **kwargs,
    )
    logger.debug(result.output)
    if result.exception:
        exc_info = result.exc_info
        traceback.print_exception(*exc_info)  # type: ignore
    if result.exit_code != 0:
        elems = helpers.recursive_str(data=args)
        elems = " ".join(elems)
        logger.error(
            f"Failed : {cli_runner.get_default_prog_name(called_cli)} {elems}"
        )  # pylint:disable=logging-fstring-interpolation
    assert result.exit_code == 0
    return result.output.rstrip()



#!/usr/bin/env python3
"""Main module, import commands and start CLI"""
import logging

from click_skeleton import helpers, version_checker
from examples.cli import PROG_NAME, __version__, main_cli

logger = logging.getLogger(__name__)


def main() -> None:
    """click skeleton entrypoint, it will check if a new version is available"""
    _ = helpers.raise_limits()
    version_check = version_checker.VersionCheckerThread(
        prog_name=PROG_NAME,
        current_version=__version__,
    )
    try:
        main_cli.main(prog_name=PROG_NAME, standalone_mode=False)
        version_check.print()
    except Exception as error:
        raise error


if __name__ == "__main__":
    main()



"""Defaults settings"""

DEFAULT_CONTEXT_SETTINGS = {
    "max_content_width": 140,
    "terminal_width": 140,
    "help_option_names": ["-h", "--help"],
    # 'help_headers_color': 'yellow',
    # 'help_options_color': 'green',
}



"""Tests VersionChecker class"""

from beartype import beartype

from click_skeleton.version_checker import VersionCheckerThread


@beartype
def test_version_checker() -> None:
    """Test that version checker is working, as this an example CLI, it will print nothing"""
    version_checker = VersionCheckerThread(
        prog_name="click-skeleton", current_version="0.0.0"
    )
    version_checker.print()



"""Simple example of a CLI made with click-skeleton"""

import logging
from typing import Any

import click

import examples.commands
from click_skeleton import ExpandedPath, add_options, backtrace, doc, skeleton

PROG_NAME = "example-cli"
__version__ = "1.0.0"
logger = logging.getLogger(PROG_NAME)

global_example_option = click.option("--global-example", help="A global option")
group_of_options = add_options(
    click.option("--option-one", help="First option"),
    click.option("--option-two", help="Second option"),
)


@skeleton(
    name=PROG_NAME,
    version=__version__,
    auto_envvar_prefix="CLI",
    invoke_without_command=True,
)
@click.pass_context
@global_example_option
@group_of_options
def main_cli(
    ctx: click.Context, global_example: str, option_one: str, option_two: str
) -> Any:
    """Simple CLI example"""
    backtrace.hook(strip_path=False, enable_on_envvar_only=False, on_tty=False)
    logger.info(f"in main_cli: ctx = {ctx}")
    logger.info(f"in main_cli: global_example = {global_example}")
    logger.info(f"in main_cli: option_one = {option_one}")
    logger.info(f"in main_cli: option_two = {option_two}")
    ctx.obj.global_option = global_example
    ctx.obj.config = "global config storage"


@main_cli.command(short_help="Generates a README.rst", aliases=["doc"])
@click.pass_context
@click.option(
    "--output",
    help="README output format",
    type=click.Choice(["rst", "markdown"]),
    default="rst",
    show_default=True,
)
def readme(ctx: click.Context, output: str) -> None:
    """Uses gen_doc click-skeleton helper to generates a complete readme"""
    doc.readme(main_cli, ctx.obj.prog_name, ctx.obj.context_settings, output)


@main_cli.command(short_help="Generates an exception")
@click.pass_context
def abort(ctx: click.Context) -> None:
    """Generates an exception on purpose (test)"""
    print(f"Global option = {ctx.obj.global_option}")
    print(f"One storage = {ctx.obj.config}")
    raise click.ClickException("throw for test purposes")


@main_cli.command(short_help="Expand option path")
@click.option(
    "--file", help="File path which expands", type=ExpandedPath(), default="~/plop"
)
def expanded_path(file: str) -> None:
    """Command with expanded path option"""
    print(file)


main_cli.add_groups_from_package(examples.commands)
